{% extends "core_components/site_base.html" %}

{% block tabtitle %}
Simmate
{% endblock %}

{% block banner %}
{% include "core_components/header.html"%}
{% endblock %}


{% block body %}

<!-- start page title -->
<div class="row pt-3">
  <div class="col-12">
    <div class="page-title-box">
      <div class="page-title-right">
        <ol class="breadcrumb m-0">
          <li class="breadcrumb-item"><a href="{% url 'third_parties:home' %}">Third-Parties</a></li>
          <li class="breadcrumb-item active">{{ provider.source }}</li>
        </ol>
      </div>
      <h4 class="page-title">Workflow Type</h4>
    </div>
  </div>
</div>
<!-- end page title -->

<!-- Workflow name and descriptoin header -->
<section class="pt-4 jumbotron text-center">
  <div class="container">
    <h1 class="jumbotron-heading">{{ provider.source }}</h1>
    <p class="lead text-muted">{{ provider.source_long }}</p>
    <a href="{{ provider.homepage }}" target="_blank" class="btn btn-secondary btn-sm">Source Website</a>
    <a href="{{ provider.source_doi }}" target="_blank" class="btn btn-secondary btn-sm">Cite</a>
    <a href="?format=api" target="_blank" class="btn btn-primary btn-sm">API</a>
    <a href="?format=json" target="_blank" class="btn btn-primary btn-sm">JSON</a>
    <a href="https://jacksund.github.io/simmate/getting_started/access_the_database/quick_start/" target="_blank"
      class="btn btn-primary btn-sm">Python-client</a>
  </div>
</section>

<!-- Search Form -->
{% include "core_components/search_form.html" %}

<!-- Search Results -->
{% include "third_parties/search_results.html" %}

{% endblock %}