{% extends "core_components/site_base.html" %}

{% block tabtitle %}
Simmate
{% endblock %}

{% block banner %}
{% include "core_components/header.html"%}
{% endblock %}


{% block body %}

<!-- start page title -->
<div class="row pt-3">
  <div class="col-12">
    <div class="page-title-box">
      <div class="page-title-right">
        <ol class="breadcrumb">
          <li class="breadcrumb-item"><a href="{% url 'workflows' %}">workflows</a></li>
          <li class="breadcrumb-item"><a href="{% url 'workflows_by_type' workflow.name_type %}">
              {{ workflow.name_type }}
            </a></li>
          <li class="breadcrumb-item"><a href="{% url 'workflow_detail' workflow.name_type workflow.name_calculator workflow.name_preset %}">
              {{ workflow.name_preset }}</a></li>
          <li class="breadcrumb-item active">submit</li>
        </ol>
      </div>
      <h4 class="page-title">Workflow Submission</h4>
    </div>
  </div>
</div>
<!-- end page title -->

<!-- Workflow name and descriptoin header -->
<section class="pt-4 jumbotron text-center">
  <div class="container">
    <h1 class="jumbotron-heading">{{ workflow.name_full }}</h1>
  </div>
</section>

<!-- Note to read parameter docs -->
<div class="alert alert-primary" role="alert">
  <i class="dripicons-information me-2"></i>
  Read through our <a class="alert-link" href="https://jacksund.github.io/simmate/parameters/" target="_blank">our
    parameter documentation</a>
  to understand what each parameter controls and the different ways you can set it.
</div>


<!-- Submission form -->
<form action="{% url 'workflow_submit' workflow.name_type workflow.name_calculator workflow.name_preset %}" method="post"
  enctype="multipart/form-data">
  {% csrf_token %}
  {% load crispy_forms_tags %}
  {{ submission_form | crispy }}
  <input type="submit" value="Submit">
</form>


{% endblock %}

{% block extrafooter %}
{% include "core_components/datatable.html" %}
{% endblock %}