# -*- coding: utf-8 -*-

"""
This file only exists because DigitalOcean and Heroku require a setup.py
to detect that python is being used. Once they update to pyproject.toml,
this file will be removed.
"""

from setuptools import setup

setup()
